from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.signals import setting_changed
from django.dispatch import receiver
from django.utils.functional import cached_property

# Retrieved from Chrome document.featurePolicy.allowedFeatures()
# with flag "Experimental Web Platform features" turned on
FEATURE_NAMES = {
    "accelerometer",
    "ambient-light-sensor",
    "autoplay",
    "camera",
    "ch-device-memory",
    "ch-downlink",
    "ch-dpr",
    "ch-ect",
    "ch-lang",
    "ch-rtt",
    "ch-ua",
    "ch-ua-arch",
    "ch-ua-full-version",
    "ch-ua-mobile",
    "ch-ua-model",
    "ch-ua-platform",
    "ch-viewport-width",
    "ch-width",
    "document-domain",
    "document-write",
    "downloads",
    "encrypted-media",
    "execution-while-not-rendered",
    "execution-while-out-of-viewport",
    "focus-without-user-activation",
    "forms",
    "fullscreen",
    "geolocation",
    "gyroscope",
    "hid",
    "idle-detection",
    "layout-animations",
    "lazyload",
    "loading-frame-default-eager",
    "magnetometer",
    "microphone",
    "midi",
    "modals",
    "orientation-lock",
    "payment",
    "picture-in-picture",
    "pointer-lock",
    "popups",
    "presentation",
    "screen-wake-lock",
    "scripts",
    "serial",
    "sync-script",
    "sync-xhr",
    "top-navigation",
    "usb",
    "vertical-scroll",
    "xr-spatial-tracking",
}


class FeaturePolicyMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        self.header_value  # Access at setup so ImproperlyConfigured can be raised
        receiver(setting_changed)(self.clear_header_value)

    def __call__(self, request):
        response = self.get_response(request)
        value = self.header_value
        if value:
            response["Feature-Policy"] = value
        return response

    @cached_property
    def header_value(self):
        setting = getattr(settings, "FEATURE_POLICY", {})
        pieces = []
        for feature, values in sorted(setting.items()):
            if feature not in FEATURE_NAMES:
                raise ImproperlyConfigured("Unknown feature {}".format(feature))
            if isinstance(values, str):
                values = (values,)

            item = [feature]
            for value in values:
                if value == "none":
                    item.append("'none'")
                elif value == "self":
                    item.append("'self'")
                else:
                    item.append(value)
            pieces.append(" ".join(item))
        return "; ".join(pieces)

    def clear_header_value(self, setting, **kwargs):
        if setting == "FEATURE_POLICY":
            try:
                del self.header_value
            except AttributeError:
                pass



from tests.testapp.views import index

try:
    from django.urls import path

    urlpatterns = [path("", index, name="index")]
except ImportError:  # Django < 2.0
    from django.conf.urls import url

    urlpatterns = [url(r"^$", index, name="index")]



import pytest
from django.core.exceptions import ImproperlyConfigured


def test_index(client):
    resp = client.get("/")

    assert resp.status_code == 200
    assert resp.content == b"Hello World"


def test_no_setting(client):
    resp = client.get("/")

    assert "Feature-Policy" not in resp


def test_anyone_can_geolocate(client, settings):
    settings.FEATURE_POLICY = {"geolocation": "*"}

    resp = client.get("/")

    assert resp["Feature-Policy"] == "geolocation *"


def test_anyone_can_geolocate_list(client, settings):
    settings.FEATURE_POLICY = {"geolocation": ["*"]}

    resp = client.get("/")

    assert resp["Feature-Policy"] == "geolocation *"


def test_no_one_can_geolocate(client, settings):
    settings.FEATURE_POLICY = {"geolocation": "none"}

    resp = client.get("/")

    assert resp["Feature-Policy"] == "geolocation 'none'"


def test_self_can_geolocate(client, settings):
    settings.FEATURE_POLICY = {"geolocation": "self"}

    resp = client.get("/")

    assert resp["Feature-Policy"] == "geolocation 'self'"


def test_example_com_can_geolocate(client, settings):
    settings.FEATURE_POLICY = {"geolocation": "https://example.com"}

    resp = client.get("/")

    assert resp["Feature-Policy"] == "geolocation https://example.com"


def test_multiple_allowed(client, settings):
    settings.FEATURE_POLICY = {"autoplay": ["self", "https://example.com"]}

    resp = client.get("/")

    assert resp["Feature-Policy"] == "autoplay 'self' https://example.com"


def test_multiple_features(client, settings):
    settings.FEATURE_POLICY = {
        "accelerometer": "self",
        "geolocation": ["self", "https://example.com"],
    }

    resp = client.get("/")

    assert (
        resp["Feature-Policy"]
        == "accelerometer 'self'; geolocation 'self' https://example.com"
    )


def test_unknown_feature(client, settings):
    settings.FEATURE_POLICY = {"accelerometor": "self"}

    with pytest.raises(ImproperlyConfigured):
        client.get("/")


def test_setting_changing(client, settings):
    settings.FEATURE_POLICY = {}
    client.get("/")  # Forces middleware instantiation
    settings.FEATURE_POLICY = {"geolocation": "self"}

    resp = client.get("/")

    assert resp["Feature-Policy"] == "geolocation 'self'"


def test_other_setting_changing(client, settings):
    settings.FEATURE_POLICY = {"geolocation": "self"}
    client.get("/")  # Forces middleware instantiation
    settings.SECRET_KEY = "foobar"

    resp = client.get("/")

    assert resp["Feature-Policy"] == "geolocation 'self'"



DEBUG = False

SECRET_KEY = "not-secret"

DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "LOCATION": ":memory:"}
}

ALLOWED_HOSTS = ["*"]

INSTALLED_APPS = []

MIDDLEWARE = [
    "django_feature_policy.FeaturePolicyMiddleware",
    "django.middleware.common.CommonMiddleware",
]

ROOT_URLCONF = "tests.testapp.urls"

STATIC_URL = "/static/"

TEMPLATES = []
